from jinja2 import Environment, FileSystemLoader
import sys, os, datetime
import json
from qualifyr.utility import include_file


def create_html_report(json_data, output_dir = '.', extra_columns = [], title = "Overall Quality Report", subtitle = ""):
    root = os.path.dirname(os.path.abspath(__file__))
    templates_dir = os.path.join(root, 'html_report', 'templates')
    env = Environment( loader = FileSystemLoader(templates_dir) )
    env.globals['include_file'] = include_file
    env.globals['os'] = os
    template = env.get_template('main.html')

    column_data =  '{ title: "Sample Name", data: "sample_name"}'
    if len(extra_columns) > 0:
        column_data = column_data + ', ' + ','.join(
            ['{{title: "{0}", data: "checks.{1}.{0}.metric_value" }}'.format(column['metric_name'], column['quality_file']) for column in extra_columns]
        )
    column_data = column_data + ', { title: "Result", data: "result" }'

    
    
    filename = os.path.join(output_dir, 'qualifyr_report.html')
    with open(filename, 'w') as fh:
        fh.write(template.render(
            asset_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'html_report', 'assets'),
            json_data = json.dumps(json_data),
            title = title,
            subtitle = subtitle,
            report_time = datetime.datetime.now().strftime("%H:%M %A %d-%m-%Y"),
            sample_numbers = len(json_data),
            num_of_pass_samples = len([sample for sample in json_data if sample['result'] == 'PASS']),
            num_of_warning_samples = len([sample for sample in json_data if sample['result'] == 'WARNING']),
            num_of_failure_samples = len([sample for sample in json_data if sample['result'] == 'FAILURE']),
            column_data = column_data,
            sort_column = len(extra_columns) + 2 # 2 default columns + extra columns
        ))
    


import sys
import json
from collections import OrderedDict
from qualifyr.utility import get_logger
from qualifyr.quality_check import QualityCheck

logger = get_logger(__file__)
class QualityFile:
    file_type = None

    def __init__(self, file_path, file_id = None):
        self.file_path = file_path 
        self.metrics = {} # a dict of metrics key is the metric name, value is the metric value
        self.quality_checks = [] # A list of tuples where the first item is the metric name, second is the value, third if the reason for failure, and the last is the failure category (PASS, FAILURE or WARNING)
        self.file_id = file_id # used in cases where there are two files for a pair of fastqs, e.g FASTQC
        self.extra_info = None
        self.overall_qc_check_result_operator = 'OR'
        
        self.parse()
    
    @property
    def type_and_id(self):
        if self.file_id:
            return '{0} {1}'.format(self.file_type, self.file_id)
        else:
            return self.file_type

    @property
    def failed_checks(self):
        return [ quality_check for quality_check in self.quality_checks if quality_check.result != 'PASS' ]

    def parse(self):
        raise NotImplementedError

    def validate(self):
        raise NotImplementedError


    def check(self, conditions): # method to check if the metrics contained in the quality file meet the conditions specified in the yml conditions file
        # Steps as follows
        # check all condition keys exist
        # loop through conditions and apply condition, one of gt,lt,lt_or_gt,gt_and_lt,eq,ne
        # return conditions that fail

        # check all condition keys exist - extract conditions for file type
        try:
            conditions_for_file_type = conditions[self.file_type]
        except KeyError as e:
            logger.error("No such quality file type {0} in conditions. The available file types in the supplied condition file are {1}".format(e, ", ".join(conditions.keys())))
            sys.exit(1)

        # loop through conditions and apply condition, one of gt,lt,lt_or_gt,gt_and_lt,eq,ne
        for metric_name in conditions_for_file_type:
            self.quality_checks.append(self.check_metric(metric_name, conditions_for_file_type))


    def check_metric(self, metric_name, conditions):
        """
        checks a metric against a condition and returns a tuple with PASS, FAILURE or WARNING and a string with a summary of the check
        Params:
            metric_name (string): name of the metric
            conditions (dict): dict containing checks to be made
        Returns:
            tuple: metric_name, metric_value, reason for warning/failure, check_result of failure
        """
        try:
            metric_value = self.metrics[metric_name]
            return QualityCheck(metric_name, metric_value, conditions)
            
        except KeyError as e:
            logger.error("No such metric {0}. The available metrics are {1}".format(e, ", ".join(self.metrics.keys())))
            sys.exit(1)

    def quality_checks_to_dict(self, only_failed_checks = True):
        quality_checks_dict = { self.type_and_id : OrderedDict() }
        if only_failed_checks:
            included_quality_checks = self.failed_checks
        else:
            included_quality_checks = self.quality_checks

        for quality_check in included_quality_checks:
            quality_checks_dict[self.type_and_id][quality_check.metric_name] = OrderedDict([
                        ('metric_value' , quality_check.metric_value),
                        ('check' , quality_check.check),
                        ('check_result' , quality_check.result)
                    ])
        return quality_checks_dict

    def output_quality_checks_string(self, output_format = 'tsv', only_failed_checks = True):
        if output_format == 'tsv':
            if only_failed_checks:
                included_quality_checks = self.failed_checks
            else:
                included_quality_checks = self.quality_checks

            checks_string = '\n'.join(
                    [ '\t'.join([
                        self.type_and_id,
                        str(quality_check.metric_name),
                        str(quality_check.metric_value),
                        str(quality_check.check),
                        str(quality_check.result)]
                        )
                        for quality_check in included_quality_checks
                    ]
                )
            return checks_string

        elif output_format == 'json':
            return json.dumps(self.quality_checks_to_dict(only_failed_checks), indent = 2 )

    def overall_qc_check_result(self):
        operator = self.overall_qc_check_result_operator
        failure_categories = [quality_check.result for quality_check in self.quality_checks]
        if operator == 'OR' and any([failure_category == 'FAILURE' for failure_category in failure_categories]):
            return 'FAILURE'
        elif operator == 'AND' and all([failure_category == 'FAILURE' for failure_category in failure_categories]):
            return 'FAILURE'
        elif operator == 'OR' and any([failure_category == 'WARNING' for failure_category in failure_categories]):
            return 'WARNING'
        elif operator == 'AND' and all([failure_category == 'WARNING' for failure_category in failure_categories]):
            return 'WARNING'
        else:
            return 'PASS'
    
    @staticmethod
    def check_multiple_files(quality_files, conditions):
        # check multiple quality files
        for quality_file in quality_files:
            quality_file.check(conditions)
        return quality_files
    
    @staticmethod
    def multiple_overall_qc_check_result(quality_files):
        failure_categories = [quality_file.overall_qc_check_result() for quality_file in quality_files]
        if 'FAILURE' in failure_categories:
            return 'FAILURE'
        elif 'WARNING' in failure_categories:
            return 'WARNING'
        else:
            return 'PASS'

    @staticmethod
    def multiple_qc_result_string(sample_name, quality_files, output_format = 'tsv', only_failed_checks = True):
        """
        Function to produce a string of the results from multiple quality check files
        """ 
        if output_format == 'tsv':
            header_string = 'file_type\tmetric_name\tmetric_value\tcheck\tcheck result\n'
            quality_check_strings = [quality_file.output_quality_checks_string(output_format, only_failed_checks) for quality_file in quality_files]
            # remove empty strings
            quality_check_strings = [quality_check_string for quality_check_string in quality_check_strings if quality_check_string != ""]
            body_string = '\n'.join(quality_check_strings)
            return header_string + body_string + '\n'
        elif output_format == 'json':
            checks_dict = {}
            extra_info_dict = OrderedDict()
            for quality_file in quality_files:
                checks_dict.update(quality_file.quality_checks_to_dict(only_failed_checks))
                extra_info_dict[quality_file.type_and_id] = quality_file.extra_info

            output_dict = OrderedDict([
                ('sample_name', sample_name),
                ('result', QualityFile.multiple_overall_qc_check_result(quality_files)),
                ('checks', checks_dict),
                ('extra_info', extra_info_dict)
            ])
            return json.dumps(output_dict, indent = 2 )


            


#!/usr/bin/env python3
import os, yaml, sys, glob, json
import argparse
from qualifyr.utility import get_logger
from qualifyr.quality_file import QualityFile
from qualifyr.quast_file import QuastFile
from qualifyr.fastqc_summary_file import FastqcFile
from qualifyr.confindr_file import ConFindrFile
from qualifyr.bactinspector_file import BactinspectorFile
from qualifyr.file_size_check_file import FileSizeCheckFile
from qualifyr.conditions_file import import_yaml_file
from qualifyr.html_report import create_html_report
from qualifyr.text_report import create_text_report

class ParserWithErrors(argparse.ArgumentParser):
    logger = get_logger(__file__)
    def error(self, message):
        self.logger.error('{0}\n\n'.format(message))
        self.print_help()
        raise Exception('')

    def is_valid_file(self, parser, arg):
        if not os.path.isfile(arg):
            parser.error("The file %s does not exist!" % arg)
        else:
            return arg

def parse_arguments():
    description = """
    A package to check quality files and assess overall pass/fail
    """
    parser = ParserWithErrors(description = description)
    subparsers = parser.add_subparsers(
            help='The following commands are available. Type qualifyr <COMMAND> -h for more help on a specific commands',
            dest='command'
        )
    
    subparsers.required = True

    # check sub command
    check_command = subparsers.add_parser('check', help='Check multiple quality metric files based on conditions and produce overall result')

    check_command.add_argument("-q", "--quast_file",
                        help="quast file path",
                        type=lambda x: parser.is_valid_file(parser, x))
    check_command.add_argument("-f", "--fastqc_file",
                        help="fastqc summary file path",
                        type=lambda x: parser.is_valid_file(parser, x),
                        nargs='+')
    check_command.add_argument("-c", "--confindr_file",
                        help="confindr report file path",
                         type=lambda x: parser.is_valid_file(parser, x))
    check_command.add_argument("-b", "--bactinspector_file",
                        help="bactinspector output file path",
                         type=lambda x: parser.is_valid_file(parser, x))
    check_command.add_argument("-z", "--file_size_check_file",
                        help="simple file size check file with headings file and size (in Mb)",
                         type=lambda x: parser.is_valid_file(parser, x))
    check_command.add_argument("-y", "--conditions_yaml_file", required=True,
                        help="conditions yaml file path",
                        type=lambda x: parser.is_valid_file(parser, x))
    check_command.add_argument("-j", "--json_output_format",
                        help="Output the check results as JSON rather than TSV (default)",
                        action='store_true')
    check_command.add_argument("-s", "--sample_name",
                        help="The name of the sample from which the quality files are derived",
                        required = True)
    check_command.add_argument("-o", "--output_dir", help="Path to output directory. If specified output will be written to file in format sample_name.qualifyr.{tsv,json}")

    
    # report sub command
    report_command = subparsers.add_parser('report', help='Produce a html report based on the qualifyr output from multiple samples')
    report_command.add_argument("-i", "--input_dir", required=True, help="Path to input directory containing multiple qualifyr json outputs")
    report_command.add_argument("-o", "--output_dir", help="Path to output directory")
    report_command.add_argument(
        "-c", "--extra_columns",
        help="Extra columns to add to the report provided as a quote enclosed comma separated list e.g 'quast.N50,quast.# contigs (>= 1000 bp),confindr.contam_status'"
    )
    report_command.add_argument("-t", "--report_title", help="Title for the report", default="Overall Quality Report")
    report_command.add_argument("-s", "--report_subtitle", help="Sub title for the report", default="")

    return parser

def run_check_command(parser, args):
    # load conditions
    conditions = import_yaml_file(args.conditions_yaml_file)
    
    # make list of quality files
    quality_files = []
    # add quast file if present
    if args.quast_file:
        quast_file = QuastFile(args.quast_file)
        quality_files.append(quast_file)
    # add fastqc file if present
    if args.fastqc_file:
        for index, fastqc_file_path in enumerate(args.fastqc_file):
            fastqc_file = FastqcFile(fastqc_file_path, index + 1)
            quality_files.append(fastqc_file)
    # add confindr file if present
    if args.confindr_file:
        confindr_file = ConFindrFile(args.confindr_file)
        quality_files.append(confindr_file)
    # add bactinspector file if present
    if args.bactinspector_file:
        bactinspector_file = BactinspectorFile(args.bactinspector_file)
        quality_files.append(bactinspector_file)
    # add file_size check file file if present
    if args.file_size_check_file:
        file_size_check_file = FileSizeCheckFile(args.file_size_check_file)
        quality_files.append(file_size_check_file)

    if len(quality_files) == 0:
        parser.error("You must specify at least one quality file")
    else:
        quality_files = QualityFile.check_multiple_files(quality_files, conditions)
        overall_result = QualityFile.multiple_overall_qc_check_result(quality_files)

        if args.json_output_format:
            output_format= 'json'
            only_failed_checks = False
        else:
            output_format = 'tsv'
            only_failed_checks = True
        
        output_string = QualityFile.multiple_qc_result_string(args.sample_name, quality_files, output_format = output_format, only_failed_checks = only_failed_checks )
        
        if output_format == 'tsv' and not args.output_dir:
            sys.stdout.write('{0}\n'.format(overall_result))

        if args.output_dir:
            with open(os.path.join(args.output_dir, '{0}.qualifyr.{1}'.format(args.sample_name, output_format)), 'w') as outfile:
                outfile.write(output_string)
        else:
            sys.stderr.write(output_string)

def run_report_command(parser, args):
    if not args.output_dir:
        args.output_dir = args.input_dir

    combined_sample_qualifyr_list = []
    extra_columns = []
    for json_file in glob.glob(os.path.join(args.input_dir, '*.qualifyr.json')):
        with open(json_file) as json_fh:
            combined_sample_qualifyr_list.append(json.load(json_fh))

    if args.extra_columns:
        for quality_file, metric_name in [column.split('.') for column in args.extra_columns.split(',')]:
            extra_columns.append({ 'quality_file': quality_file, 'metric_name': metric_name})
            for sample in combined_sample_qualifyr_list:
                if quality_file not in sample['checks'] or metric_name not in sample['checks'][quality_file]:
                    parser.error('Sample {0} does not have data for {1}/{2} as specified in the -c/--extra_columns argument'.format(sample['sample_name'], quality_file, metric_name))

    # make text report
    create_text_report(
        combined_sample_qualifyr_list,
        output_dir = args.output_dir
    )

    # make html report
    create_html_report(
        combined_sample_qualifyr_list,
        output_dir = args.output_dir,
        extra_columns = extra_columns,
        title = args.report_title,
        subtitle = args.report_subtitle
    )

def choose_command(parser, args):
    if args.command == 'check':
        run_check_command(parser, args)
    elif args.command == 'report':
        run_report_command(parser, args)

def main():
    parser = parse_arguments()
    args = parser.parse_args()
    choose_command(parser, args)






import yaml, sys
from qualifyr.utility import get_logger

logger = get_logger(__file__)
def check_conditions(conditions):
    for file_type in conditions:
        for metric in conditions[file_type]:
            if 'failure' not in conditions[file_type][metric] and 'warning' not in conditions[file_type][metric]:
                logger.error('ERROR in conditions yml file')
                logger.error('Every metric in the conditions file must have a warning or failure condition')
                sys.exit(1)



def import_yaml_file(filepath):
    with open(filepath) as yml_file:
        conditions = yaml.load(yml_file.read())
    check_conditions(conditions)
    return conditions


import csv, os

def get_headers(sample_data):
    """
    get headers based on keys in the combined qualifyr json data
    Returns a list of header columns sample_name,result,quality_file.metric<1..N>
    """
    header_items = ['sample_name', 'result']
    for quality_file in sorted(sample_data['checks'].keys()):
        for metric in sorted(sample_data['checks'][quality_file].keys()):
            for sub_metric in ['metric_value', 'check_result']:
                header_items.append(f'{quality_file}.{metric}.{sub_metric}')
    return header_items

def create_text_report(json_data, output_dir = '.'):
    """
    Create a TSV file with one row per sample
    """
    header_items = get_headers(json_data[0])

    filename = os.path.join(output_dir, 'qualifyr_report.tsv')
    with open(filename , 'w') as fh:
        writer = csv.writer(fh, delimiter='\t')
        writer.writerow(header_items)
        for sample_data in sorted(json_data, key=lambda sample_data: sample_data['sample_name']):
            row_items = []
            for index, header_item in enumerate(header_items):
                if index < 2:
                    row_items.append(sample_data[header_item])
                else:
                    quality_file, metric, sub_metric = header_item.split('.')
                    row_items.append(sample_data['checks'][quality_file][metric][sub_metric])

            writer.writerow(row_items)
    


import logging
import io
import os
import colorlog
import base64

def get_logger(filepath, level=logging.DEBUG):
    # create logger
    logger = logging.getLogger(os.path.basename(filepath))
    logger.setLevel(level)
    # create console handler and set level to debug
    console = logging.StreamHandler()
    console.setLevel(level)
    # create formatter
    format_str = '%(asctime)s - %(levelname)-8s - %(message)s'
    date_format = '%Y-%m-%d %H:%M:%S'
    if os.isatty(2):
        cformat = '%(log_color)s' + format_str
        colors = {'DEBUG': 'reset',
                  'INFO': 'reset',
                  'WARNING': 'bold_yellow',
                  'ERROR': 'bold_red',
                  'CRITICAL': 'bold_red'}
        formatter = colorlog.ColoredFormatter(cformat, date_format,
                                              log_colors=colors)
    else:
        formatter = logging.Formatter(format_str, date_format)
    # add formatter to ch
    console.setFormatter(formatter)
    # add ch to logger
    logger.addHandler(console)

    return logger

def string_to_num(string, default = None):
    try:
        return int(string)
    except ValueError:
        try :
            return float(string)
        except ValueError:
            if default:
                return default
            else:
                return string

def include_file(name, fdir=None, b64=False):
    logger = logging.getLogger(__name__)
    try:
        if fdir is None:
            fdir = ''
        if b64:
            with io.open (os.path.join(fdir, name), "rb") as f:
                return base64.b64encode(f.read()).decode('utf-8')
        else:
            with io.open (os.path.join(fdir, name), "r", encoding='utf-8') as f:
                return f.read()
    except (OSError, IOError) as e:
        logger.error("Could not include file '{}': {}".format(name, e))


class QualityCheck:
    def __init__(self, metric_name, metric_value, conditions):
        self.metric_name = metric_name
        self.metric_value = metric_value
        self.check = None
        self.result = None

        self.check_metric(conditions)


    
    def check_metric(self, conditions):
        for check_result in ['failure', 'warning']:
            if check_result in conditions[self.metric_name]:
                condition_type = conditions[self.metric_name][check_result]['condition_type']
                condition_value = conditions[self.metric_name][check_result]['condition_value']
                # possible conditions
                # greater than
                if condition_type == 'gt':
                    self.check = '> {0}'.format(condition_value)
                    if self.metric_value > condition_value:
                        self.result = check_result.upper()
                        return
                # less than
                elif condition_type == 'lt':
                    self.check = '< {0}'.format(condition_value)
                    if self.metric_value < condition_value:
                        self.result = check_result.upper()
                        return
                # less than or greater than
                elif condition_type == 'lt_or_gt':
                    self.check = '< {0} or > {1}'.format(condition_value[0], condition_value[1])
                    if (self.metric_value < condition_value[0] or self.metric_value > condition_value[1]):
                        self.result = check_result.upper()
                        return
                # less than or greater than
                elif condition_type == 'gt_and_lt':
                    self.check = '> {0} and < {1}'.format(condition_value[0], condition_value[1])
                    if (self.metric_value > condition_value[0] and self.metric_value < condition_value[1]):
                        self.result = check_result.upper()
                        return
                # equal to
                elif condition_type == 'eq':
                    self.check = 'Equals {0}'.format(condition_value)
                    if self.metric_value == condition_value:
                        self.result = check_result.upper()
                        return
                # less than and greater than
                elif condition_type == 'ne':
                    self.check = 'Does not equal {0}'.format(condition_value)
                    if self.metric_value != condition_value:
                        self.result = check_result.upper()
                        return
                # any
                elif condition_type == 'any':
                    self.check = 'One of {0}'.format(', '.join(condition_value))
                    if self.metric_value in condition_value:
                        self.result = check_result.upper()
                        return
        self.result = 'PASS'
        return
    
    def to_tuple(self):
        return (self.metric_name, self.metric_value, self.check, self.result)


import csv, sys
from qualifyr.quality_file import QualityFile
from qualifyr.utility import string_to_num, get_logger

class QuastFile(QualityFile):
    logger = get_logger(__file__)
    file_type = 'quast'

    def validate(self):
        # method to check file looks like what it says it is
        '''Returns valid rows from file. An empty list if invalid'''
        with open(self.file_path) as fh:
            reader = csv.reader(fh, delimiter="\t", quotechar='"')
            rows = list(reader)
            first_row = rows[0]
            last_row = rows[-1]
            if first_row[0] == 'Assembly' and last_row[0] == '# N\'s per 100 kbp':
                return rows[1:]
            else:
                return []
    

    def parse(self):
        # read in file and make a dict:
        valid_rows = self.validate()
        if len(valid_rows) == 0:
            self.logger.error('{0} file invalid'.format(self.file_type))
            raise(Exception)

        else:
            for row in valid_rows:
                self.metrics[row[0]] = string_to_num(row[1])



